# :coding: utf-8

import getpass
import os
import socket

import pytest
import toml

import wiz.config


@pytest.fixture(autouse=True)
def environment(mocker, monkeypatch):
    """Mock environment to use when fetching config."""
    mocker.patch.object(socket, "gethostname", return_value="__HOSTNAME__")
    mocker.patch.object(getpass, "getuser", return_value="__USER__")
    mocker.patch.object(os.path, "expanduser", return_value="__HOME__")


@pytest.fixture()
def personal_plugin(mocker, temporary_directory):
    """Mock personal plugin path."""
    path = os.path.join(temporary_directory, ".wiz", "plugins")
    mocker.patch.object(os.path, "expanduser", return_value=temporary_directory)
    os.makedirs(path)

    plugin_path = os.path.join(path, "plugin.py")
    with open(plugin_path, "w") as stream:
        stream.write(
            "IDENTIFIER = \"plugin1\"\n"
            "\n"
            "def register(config):\n"
            "    config[\"KEY1\"] = \"VALUE1\"\n"
        )

    return plugin_path


@pytest.fixture()
def personal_configuration(mocker, temporary_directory):
    """Mock personal configuration."""
    data = {
        "registry": {"paths": ["/registry1"]},
        "environ": {"initial": {"ENVIRON_TEST1": "VALUE"}}
    }

    path = os.path.join(temporary_directory, ".wiz")
    mocker.patch.object(os.path, "expanduser", return_value=temporary_directory)
    os.makedirs(path)

    config_path = os.path.join(path, "config.toml")
    with open(config_path, "w") as stream:
        toml.dump(data, stream)

    return config_path


@pytest.fixture()
def mocked_discover_plugins(mocker):
    """Return mocked 'wiz.config._discover_plugins' function."""
    return mocker.patch.object(wiz.config, "_discover_plugins", return_value=[])


@pytest.mark.usefixtures("mocked_discover_plugins")
def test_fetch():
    """Fetch configuration."""
    config = wiz.config.fetch(refresh=True)

    assert config == {
        "registry":  {"paths": []},
        "environ": {
            "initial": {},
            "passthrough": []
        },
        "resolver": {
            "maximum_combinations": 10,
            "maximum_attempts": 15,
        },
        "command": {
            "max_content_width": 90,
            "verbosity": "info",
            "no_local": False,
            "no_cwd": False,
            "ignore_implicit": False,
            "list": {
                "package": {
                    "all": False,
                    "no_arch": False
                },
                "command": {
                    "all": False,
                    "no_arch": False
                },
            },
            "search": {
                "all": False,
                "no_arch": False,
                "type": "all"
            },
            "view": {
                "json_view": False
            },
            "freeze": {
                "format": "wiz"
            },
            "install": {
                "overwrite": False
            },
            "edit": {
                "overwrite": False
            },
            "analyze": {
                "no_arch": False,
                "verbose": False,
            }
        }
    }

    # The previous config is returned without forcing a 'refreshed' config.
    del config["command"]
    assert wiz.config.fetch().get("command") is None
    assert wiz.config.fetch(refresh=True).get("command") is not None


@pytest.mark.usefixtures("mocked_discover_plugins")
@pytest.mark.usefixtures("personal_configuration")
def test_fetch_with_personal():
    """Fetch configuration with personal configuration."""
    config = wiz.config.fetch(refresh=True)

    assert config == {
        "registry":  {"paths": ["/registry1"]},
        "environ": {
            "initial": {
                "ENVIRON_TEST1": "VALUE"
            },
            "passthrough": []
        },
        "resolver": {
            "maximum_combinations": 10,
            "maximum_attempts": 15,
        },
        "command": {
            "max_content_width": 90,
            "verbosity": "info",
            "no_local": False,
            "no_cwd": False,
            "ignore_implicit": False,
            "list": {
                "package": {
                    "all": False,
                    "no_arch": False
                },
                "command": {
                    "all": False,
                    "no_arch": False
                },
            },
            "search": {
                "all": False,
                "no_arch": False,
                "type": "all"
            },
            "view": {
                "json_view": False
            },
            "freeze": {
                "format": "wiz"
            },
            "install": {
                "overwrite": False
            },
            "edit": {
                "overwrite": False
            },
            "analyze": {
                "no_arch": False,
                "verbose": False,
            }
        }
    }

    # The previous config is returned without forcing a 'refreshed' config.
    del config["command"]
    assert wiz.config.fetch().get("command") is None
    assert wiz.config.fetch(refresh=True).get("command") is not None


@pytest.mark.usefixtures("mocked_discover_plugins")
def test_fetch_error(logger, personal_configuration):
    """Fail to fetch a configuration."""
    with open(personal_configuration, "w") as stream:
        stream.write("incorrect")

    wiz.config.fetch(refresh=True)

    logger.warning.assert_called_once_with(
        "Failed to load configuration from \"{0}\" [Key name found without "
        "value. Reached end of file. (line 1 column 10 char 9)]"
        .format(personal_configuration)
    )


def test_fetch_with_plugin(mocker, mocked_discover_plugins):
    """Register plugin after fetching config."""
    def _register(_config):
        """Register config"""
        _config["registry"] = "__NEW_REGISTRY__"

    plugin = mocker.Mock(register=_register)
    mocked_discover_plugins.return_value = [plugin]

    config = wiz.config.fetch(refresh=True)
    assert config["registry"] == "__NEW_REGISTRY__"


def test_fetch_with_plugin_error(logger, mocked_discover_plugins):
    """Fail to register a plugin."""
    mocked_discover_plugins.return_value = [wiz]

    wiz.config.fetch(refresh=True)

    assert logger.warning.call_count == 1
    args, _ = logger.warning.call_args
    assert (
       "Failed to register plugin from \"{0}\""
       .format(wiz.__file__)
    ) in args[0]


def test_discover_plugins(mocker):
    """Discover and return plugins."""
    plugins = wiz.config._discover_plugins()

    assert len(plugins) == 2
    assert sorted([p.IDENTIFIER for p in plugins]) == ["environ", "installer"]

    config = {}
    for plugin in plugins:
        plugin.register(config)

    assert config == {
        "callback": {
            "install": mocker.ANY
        },
        "environ": {
            "initial": {
                "HOME": mocker.ANY,
                "HOSTNAME": "__HOSTNAME__",
                "USER": "__USER__",
                "PATH": (
                    "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:"
                    "/sbin:/bin"
                )
            }
        }
    }


@pytest.mark.usefixtures("personal_plugin")
def test_discover_plugins_with_personal(mocker):
    """Discover and return plugins with personal plugin path."""
    plugins = wiz.config._discover_plugins()

    assert len(plugins) == 3
    assert sorted([p.IDENTIFIER for p in plugins]) == [
        "environ", "installer", "plugin1"
    ]

    config = {}
    for plugin in plugins:
        plugin.register(config)

    assert config == {
        "callback": {
            "install": mocker.ANY
        },
        "environ": {
            "initial": {
                "HOME": mocker.ANY,
                "HOSTNAME": "__HOSTNAME__",
                "USER": "__USER__",
                "PATH": (
                    "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:"
                    "/sbin:/bin"
                )
            }
        },
        "KEY1": "VALUE1"
    }


def test_discover_plugin_error(logger, personal_plugin):
    """Fail to register a plugin."""
    with open(personal_plugin, "w") as stream:
        stream.write("incorrect")

    wiz.config.fetch(refresh=True)

    logger.warning.assert_called_once_with(
        "Failed to load plugin from \"{0}\" [name 'incorrect' is not defined]"
        .format(personal_plugin)
    )



# :coding: utf-8

from __future__ import print_function

import datetime
import getpass
import io
import os
import sys
import tempfile

import colorama
import pystache
import sawmill
import sawmill.compatibility
import sawmill.filterer.item
import sawmill.filterer.level
import sawmill.formatter.field
import sawmill.formatter.mustache
import sawmill.handler.stream
import sawmill.logger.classic

import wiz.filesystem

#: Top level handler responsible for relaying all logs to other handlers.
root = sawmill.root

#: Log levels ordered by severity. Do not rely on the index of the level name
# as it may change depending on the configuration.
levels = sawmill.levels


def configure():
    """Configure logging handlers.

    A file handler is created to log warnings and greater to :file:`wiz/logs`
    under system temporary directory.

    .. note::

        Standard Python logging are redirected to :mod:`sawmill` to unify
        the logging systems.

    """
    # Stderr handler
    stderr_handler = sawmill.handler.stream.Stream(sys.stderr)
    stderr_formatter = Formatter(
        "{{level}}: {{message}}{{#traceback}}\n{{.}}:{{/traceback}}\n"
    )
    stderr_handler.formatter = stderr_formatter

    stderr_filterer = sawmill.filterer.level.Level(min="info", max=None)
    stderr_handler.filterer = stderr_filterer

    # File handler
    logging_path_prefix = os.path.join(tempfile.gettempdir(), "wiz", "logs")
    wiz.filesystem.ensure_directory(logging_path_prefix)

    pid = os.getpid()
    timestamp = datetime.datetime.now().strftime("%Y%m%d")

    file_path = os.path.join(
        logging_path_prefix, "{}_{}.log".format(pid, timestamp)
    )
    file_filterer = sawmill.filterer.level.Level(min="warning", max=None)
    file_stream = open(file_path, "a", 1)
    file_handler = sawmill.handler.stream.Stream(file_stream)
    file_handler.filterer = file_filterer

    file_formatter = sawmill.formatter.field.Field([
        "date", "*", "name", "level", "message", "traceback"
    ])
    file_handler.formatter = file_formatter

    sawmill.root.handlers = {
        "stderr": stderr_handler,
        "file": file_handler
    }

    # Configure standard Python logging to redirect all logs to sawmill for
    # handling here. This unifies the logging systems.
    sawmill.compatibility.redirect_standard_library_logging()


def configure_for_debug():
    """Configure single error handler for debugging purpose.

    A standard error handler is created to log errors and greater into to a
    :class:`io.StringIO` instance.

    A standard waring handler is created to log warnings into to a
    :class:`io.StringIO` instance.

    A tuple with the two :class:`io.StringIO` instances created is returned.

    """
    error_captured = io.StringIO()

    # Error handler
    error_handler = sawmill.handler.stream.Stream(error_captured)
    error_formatter = Formatter(
        "{{level}}: {{message}}{{#traceback}}\n{{.}}:{{/traceback}}\n",
        with_color=False
    )
    error_handler.formatter = error_formatter

    error_filterer = sawmill.filterer.level.Level(min="error", max=None)
    error_handler.filterer = error_filterer

    warning_captured = io.StringIO()

    # Warning handler
    warning_handler = sawmill.handler.stream.Stream(warning_captured)
    warning_formatter = Formatter("{{level}}: {{message}}\n", with_color=False)
    warning_handler.formatter = warning_formatter

    warning_filterer = sawmill.filterer.level.Level(
        min="warning", max="warning"
    )
    warning_handler.filterer = warning_filterer

    sawmill.root.handlers = {
        "error": error_handler,
        "warning": warning_handler
    }

    return error_captured, warning_captured


class Formatter(sawmill.formatter.mustache.Mustache):
    """:term:`Mustache` template to format :class:`logs <sawmill.log.Log>`.
    """
    #: Color symbols per level.
    _COLOR = {
        "error": colorama.Fore.RED,
        "err": colorama.Fore.RED,
        "warning": colorama.Fore.YELLOW,
        "warn": colorama.Fore.YELLOW,
        "info": colorama.Fore.CYAN,
        "except": colorama.Style.BRIGHT + colorama.Fore.RED,
    }

    def __init__(self, template, with_color=True):
        """Initialize with :term:`Mustache` template."""
        self._renderer = pystache.Renderer(escape=lambda value: value)
        self._with_color = with_color
        super(Formatter, self).__init__(template, batch=False)

    def format(self, logs):
        """Format *logs* for display."""
        data = []

        for log in logs:
            line = self._renderer.render(self.template, log)

            if self._with_color and "level" in log.keys():
                if log["level"] in self._COLOR.keys():
                    line = (
                        self._COLOR[log["level"]] + line +
                        colorama.Style.RESET_ALL
                    )

            data.append(line)

        return data


class Logger(sawmill.logger.classic.Classic):
    """Extended logger with timestamp and username information."""

    def prepare(self, *args, **kw):
        """Prepare and return a log for emission."""
        log = super(Logger, self).prepare(*args, **kw)

        if "username" not in log:
            log["username"] = getpass.getuser()

        if "date" not in log:
            log["date"] = datetime.datetime.now().strftime("%Y_%m_%d-%H_%M_%S")

        return log

    def debug_traceback(self):
        """Log a traceback as debug level message."""
        self.debug("Traceback for latest error", traceback=True)



# :coding: utf-8

import traceback


class WizError(Exception):
    """Base class for Wiz specific errors."""

    def __init__(self, message):
        """Initialize with *message*.

        :param message: Message describing the issue.

        """
        self.message = message
        self.traceback = traceback.format_exc()

    def __str__(self):
        """Return human readable representation."""
        return str(self.message)

    def __eq__(self, other):
        """Compare with *other*."""
        if isinstance(other, WizError):
            return self.message == other.message
        return False


class CurrentSystemError(WizError):
    """Raise when the system is incorrect."""

    def __init__(self, message):
        """Initialize with *message*.

        :param message: Message describing the issue.

        """
        super(CurrentSystemError, self).__init__(message=message)


class UnsupportedPlatform(CurrentSystemError):
    """Raise when the current platform is not supported."""

    def __init__(self, platform):
        """Initialize with *platform*.

        :param platform: Lowercase name of the current platform as returned by
            :func:`platform.system`.

        """
        super(UnsupportedPlatform, self).__init__(
            message="The current platform is not supported: {}".format(platform)
        )


class RequestNotFound(WizError):
    """Raise when a request cannot be processed."""

    def __init__(self, message):
        """Initialize with *message*.

        :param message: Message describing the issue.

        """
        super(RequestNotFound, self).__init__(message=message)


class DefinitionError(WizError):
    """Raise when a definition is incorrect."""

    def __init__(self, message):
        """Initialize with *message*.

        :param message: Message describing the issue.

        """
        super(DefinitionError, self).__init__(message=message)


class PackageError(WizError):
    """Raise when a package is incorrect."""

    def __init__(self, message):
        """Initialize with *message*.

        :param message: Message describing the issue.

        """
        super(PackageError, self).__init__(message=message)


class RequirementError(WizError):
    """Raise when a requirement is incorrect."""

    def __init__(self, message):
        """Initialize with *message*.

        :param message: Message describing the issue.

        """
        super(RequirementError, self).__init__(message=message)


class VersionError(WizError):
    """Raise when a version is incorrect."""

    def __init__(self, message):
        """Initialize with *message*.

        :param message: Message describing the issue.

        """
        super(VersionError, self).__init__(message=message)


class GraphResolutionError(WizError):
    """Raise when the graph is incorrect."""

    def __init__(self, message):
        """Initialize with *message*.

        :param message: Message describing the issue.

        """
        super(GraphResolutionError, self).__init__(message=message)


class GraphConflictsError(GraphResolutionError):
    """Raise when unsolvable conflicts are found in the graph."""

    def __init__(self, conflicting):
        """Initialize with a list of conflict mappings.

        :param conflicting: Mapping of conflicting node identifiers per
            requirement. It should be in the form of::

                {
                    Requirement("foo >=0.1.0, <1"): {"bar", "bim"},
                    Requirement("foo >2"): {"baz},
                    ...
                }

        """
        # Sort conflicting requirements per ascending number of conflicting
        # nodes so that resolver attempts to downgrade versions of the smallest
        # cluster first.
        self.conflicts = sorted(
            conflicting.items(), key=lambda t: (len(t[1]), str(t[0]))
        )

        def _format(requirement, identifiers):
            """Display conflicting *mapping*."""
            # Convert set to list and sort to make message deterministic.
            identifiers = sorted(identifiers)

            # Truncate list if too long.
            if len(identifiers) > 3:
                _ellipsis = "+{} packages...".format(len(identifiers[3:]))
                identifiers = identifiers[:3] + [_ellipsis]

            return "  * {requirement} \t[{identifiers}]".format(
                requirement=requirement,
                identifiers=", ".join(identifiers)
            )

        super(GraphConflictsError, self).__init__(
            message=(
                "The dependency graph could not be resolved due to the "
                "following requirement conflicts:\n"
                "{}\n".format("\n".join([_format(*c) for c in self.conflicts]))
            )
        )

    def __eq__(self, other):
        """Compare with *other*."""
        if isinstance(other, GraphConflictsError):
            return self.conflicts == other.conflicts
        return super(GraphConflictsError, self).__eq__(other)


class GraphInvalidNodesError(GraphResolutionError):
    """Raise when invalid nodes are found in the graph."""

    def __init__(self, error_mapping):
        """Initialize with a list of conflict mappings.

        :param error_mapping: Mapping containing list of error messages per node
            identifier.

        """
        self.error_mapping = error_mapping

        super(GraphInvalidNodesError, self).__init__(
            message=(
                "The dependency graph could not be resolved due to the "
                "following error(s):\n"
                "{}\n".format(
                    "\n".join([
                        "  * {}: {}".format(identifier, error)
                        for identifier, errors in sorted(error_mapping.items())
                        for error in errors
                    ])
                )
            )
        )

    def __eq__(self, other):
        """Compare with *other*."""
        if isinstance(other, GraphInvalidNodesError):
            return self.error_mapping == other.error_mapping
        return super(GraphInvalidNodesError, self).__eq__(other)


class GraphVariantsError(GraphResolutionError):
    """Raise when division of the graph into new combinations in required."""

    def __init__(self):
        """Initialize Error."""
        super(GraphResolutionError, self).__init__(
            message="The current graph has conflicting variants."
        )


class FileExists(WizError):
    """Raise when a file already exists."""

    def __init__(self, path):
        """Initialize with *message*.

        :param path: File path.

        """
        super(FileExists, self).__init__(
            message="{!r} already exists.".format(path)
        )


class DefinitionsExist(WizError):
    """Raise when definitions already exists in a registry."""

    def __init__(self, definitions):
        """Initialize with list of existing *definitions*.

        :param definitions: List of definition labels (e.g. "'foo' [0.1.0]").

        """
        self.definitions = definitions

        super(DefinitionsExist, self).__init__(
            message="{} definition(s) already exist in registry.".format(
                len(definitions)
            )
        )


class InstallError(WizError):
    """Raise when the installation of a definition failed."""

    def __init__(self, message):
        """Initialize with *message*.

        :param message: Message describing the issue.

        """
        super(InstallError, self).__init__(message=message)


class InstallNoChanges(WizError):
    """Raise when no content was detected in a release request."""

    def __init__(self):
        """Initialize Error."""
        super(InstallNoChanges, self).__init__(message="Nothing to install.")



# :coding: utf-8

"""Common symbols."""

#: Environment variable corresponding to the 'install-location' definition key.
INSTALL_LOCATION = "INSTALL_LOCATION"

#: Environment variable corresponding to the 'install-root' definition key.
INSTALL_ROOT = "INSTALL_ROOT"

#: Separator between the normal arguments and the command to run.
COMMAND_SEPARATOR = "--"

#: Separator between namespaces and identifier.
NAMESPACE_SEPARATOR = "::"

#: Default value when a definition value is unknown.
UNSET_VALUE = "-"

#: Package request type.
PACKAGE_REQUEST_TYPE = "package"

#: Command request type.
COMMAND_REQUEST_TYPE = "command"

#: Identifier for packages which should be use implicitly in context.
IMPLICIT_PACKAGE = "implicit-packages"

#: History action for system identification.
SYSTEM_IDENTIFICATION_ACTION = "IDENTIFY_SYSTEM"

#: History action for definitions collection.
DEFINITIONS_COLLECTION_ACTION = "FETCH_DEFINITIONS"

#: History action for graph creation.
GRAPH_CREATION_ACTION = "CREATE_GRAPH"

#: History action for graph update.
GRAPH_UPDATE_ACTION = "UPDATE_GRAPH"

#: History action for graph combination extraction.
GRAPH_COMBINATION_EXTRACTION_ACTION = "EXTRACT_GRAPH_COMBINATION"

#: History action for node replacement in graph.
GRAPH_NODES_REPLACEMENT_ACTION = "REPLACE_NODES"

#: History action for computation of graph distance mapping.
GRAPH_DISTANCE_COMPUTATION_ACTION = "CREATE_DISTANCE_MAPPING"

#: History action for node creation within graph.
GRAPH_NODE_CREATION_ACTION = "CREATE_NODE"

#: History action for node removal within graph.
GRAPH_NODE_REMOVAL_ACTION = "REMOVE_NODE"

#: History action for link creation within graph.
GRAPH_LINK_CREATION_ACTION = "CREATE_LINK"

#: History action for version conflicts identification within graph.
GRAPH_VERSION_CONFLICTS_IDENTIFICATION_ACTION = "IDENTIFY_VERSION_CONFLICTS"

#: History action for variants conflicts identification within graph.
GRAPH_VARIANT_CONFLICTS_IDENTIFICATION_ACTION = "IDENTIFY_VARIANT_CONFLICTS"

#: History action for error identification within graph.
GRAPH_ERROR_IDENTIFICATION_ACTION = "IDENTIFY_ERROR"

#: History action for resolution error within graph.
GRAPH_RESOLUTION_FAILURE_ACTION = "RESOLUTION_ERROR"

#: History action for package extraction from graph.
GRAPH_PACKAGES_EXTRACTION_ACTION = "EXTRACT_PACKAGES"

#: History action for context extraction packages list.
CONTEXT_EXTRACTION_ACTION = "EXTRACT_CONTEXT"

#: History action for exception raised.
EXCEPTION_RAISE_ACTION = "RAISE_EXCEPTION"



# :coding: utf-8

import pytest
import subprocess

import wiz.spawn
import wiz.environ


@pytest.fixture()
def mocked_subprocess_call(mocker):
    """Return mocked subprocess.call function."""
    return mocker.patch.object(subprocess, "call")


@pytest.fixture()
def spied_environ_substitute(mocker):
    """Return mocked environ.substitute function."""
    return mocker.spy(wiz.environ, "substitute")


def test_execute(mocked_subprocess_call, spied_environ_substitute, logger):
    """Execute command within environment."""
    wiz.spawn.execute(["app_exe", "-v", "/script"], "__ENVIRON__")

    assert spied_environ_substitute.call_count == 3
    spied_environ_substitute.assert_any_call("app_exe", "__ENVIRON__")
    spied_environ_substitute.assert_any_call("-v", "__ENVIRON__")
    spied_environ_substitute.assert_any_call("/script", "__ENVIRON__")

    mocked_subprocess_call.assert_called_once_with(
        ["app_exe", "-v", "/script"], env="__ENVIRON__"
    )

    logger.info.assert_called_once_with("Start command: app_exe -v /script")
    logger.error.assert_not_called()
    logger.debug.assert_not_called()


def test_execute_fail(mocked_subprocess_call, spied_environ_substitute, logger):
    """Fail to execute command within environment."""
    exception = OSError("Oh Shit!")
    mocked_subprocess_call.side_effect = exception

    wiz.spawn.execute(["app_exe", "-v", "/script"], "__ENVIRON__")

    assert spied_environ_substitute.call_count == 3
    spied_environ_substitute.assert_any_call("app_exe", "__ENVIRON__")
    spied_environ_substitute.assert_any_call("-v", "__ENVIRON__")
    spied_environ_substitute.assert_any_call("/script", "__ENVIRON__")

    mocked_subprocess_call.assert_called_once_with(
        ["app_exe", "-v", "/script"], env="__ENVIRON__"
    )

    logger.info.assert_called_once_with("Start command: app_exe -v /script")
    logger.error.assert_called_once_with(
        "Executable can not be found within resolved environment [app_exe]"
    )



# :coding: utf-8

import platform as _platform

import distro

import wiz.exception
import wiz.history
import wiz.symbol
import wiz.utility

#: Operating System group mapping
OS_MAPPING = {
    "el": ["centos", "redhat"]
}


def query(platform=None, architecture=None, os_name=None, os_version=None):
    """Return system mapping.

    The mapping should be in the form of::

        {
            "platform": "linux",
            "arch": "x86_64",
            "os": {
                "name": "centos",
                "version": "7.3.161"
            }
        }

    :param platform: Indicate a platform identifier which would override the
        platform identifier queried. Default is None.

    :param architecture: Indicate an architecture identifier which would
        override the architecture identifier queried. Default is None.

    :param os_name: Indicate an operating system identifier which would override
        the operating system identifier queried. Default is None.

    :param os_version: Indicate an operating system version which would override
        the operating system version queried. Default is None.

    :return: System mapping.

    :raise: :exc:`wiz.exception.IncorrectSystem` if current system could not
        be identified or is not supported.

    """
    name = _platform.system().lower()

    try:
        if name == "linux":
            mapping = query_linux()
        elif name == "darwin":
            mapping = query_mac()
        elif name == "windows":
            mapping = query_windows()
        else:
            raise wiz.exception.UnsupportedPlatform(name)

    except wiz.exception.VersionError as error:
        raise wiz.exception.CurrentSystemError(
            "The operating system version found seems incorrect "
            "[{}]".format(error)
        )

    if platform is not None:
        mapping["platform"] = platform

    if architecture is not None:
        mapping["arch"] = architecture

    if os_name is not None:
        mapping["os"]["name"] = os_name

    if os_version is not None:
        mapping["os"]["version"] = os_version

    wiz.history.record_action(
        wiz.symbol.SYSTEM_IDENTIFICATION_ACTION, system=mapping
    )

    return mapping


def query_linux():
    """Return Linux system mapping.

    :return: Mapping in the form of
        ::

            {
                "platform": "linux",
                "arch": "x86_64",
                "os": {
                    "name": "centos",
                    "version": "7.3.161"
                }
            }

    """
    distribution, version, _ = distro.linux_distribution(
        full_distribution_name=False
    )

    return {
        "platform": "linux",
        "arch": _platform.machine(),
        "os": {
            "name": distribution,
            "version": wiz.utility.get_version(version)
        }
    }


def query_mac():
    """Return mac system mapping.

    :return: Mapping in the form of
        ::

            {
                "platform": "mac",
                "arch": "x86_64",
                "os": {
                    "name": "mac",
                    "version": "10.15.5"
                }
            }

    """
    return {
        "platform": "mac",
        "arch": _platform.machine(),
        "os": {
            "name": "mac",
            "version": wiz.utility.get_version(_platform.mac_ver()[0])
        }
    }


def query_windows():
    """Return windows system mapping.

    :return: Mapping in the form of
        ::

            {
                "platform": "windows",
                "arch": "x86_64",
                "os": {
                    "name": "windows",
                    "version": "10.0.10240"
                }
            }

    .. warning::

        The Windows versions superior to 8 will not be recognised properly with
        a Python version under 2.7.11

        https://hg.python.org/cpython/raw-file/53d30ab403f1/Misc/NEWS

        Also a bug as been introduced in Python 2.7.11 that prevent the
        recognition of old Windows version

        https://bugs.python.org/issue26513

    """
    return {
        "platform": "windows",
        "arch": _platform.machine(),
        "os": {
            "name": "windows",
            "version": wiz.utility.get_version(_platform.win32_ver()[1])
        }
    }


def validate(definition, system_mapping):
    """Validate *definition* against system *mapping*.

    :param definition: Instance of :class:`wiz.definition.Definition`.

    :param system_mapping: System mapping as returned by :func:`query`.

    :return: Boolean value.

    """
    system = definition.system

    # If no system is set on the definition, it is considered compatible with
    # any platform.
    if len(system) == 0:
        return True

    # Filter platform if necessary.
    platform_identifier = system_mapping.get("platform")
    if system.get("platform", platform_identifier) != platform_identifier:
        return False

    # Filter architecture if necessary.
    architecture = system_mapping.get("arch")
    if system.get("arch", architecture) != architecture:
        return False

    # Filter operating system version if necessary.
    os_system = system.get("os")
    if os_system is not None:
        try:
            requirement = wiz.utility.get_requirement(os_system)
        except wiz.exception.RequirementError:
            raise wiz.exception.DefinitionError(
                "The operating system requirement is incorrect: {}".format(
                    os_system
                )
            )

        os_mapping = system_mapping.get("os", {})

        if not (
            requirement.name == os_mapping.get("name") or
            os_mapping.get("name") in OS_MAPPING.get(requirement.name, [])
        ):
            return False

        if os_mapping.get("version") not in requirement.specifier:
            return False

    return True



# :coding: utf-8

import collections
import os
import shlex

import wiz.definition
import wiz.environ
import wiz.exception
import wiz.filesystem
import wiz.graph
import wiz.logging
import wiz.package
import wiz.registry
import wiz.spawn
import wiz.symbol
import wiz.system
import wiz.utility
from ._version import __version__


def fetch_definition_mapping(paths, max_depth=None, system_mapping=None):
    """Return mapping including all definitions available under *paths*.

    Mapping returned should be in the form of::

        {
            "command": {
                "foo-app": "foo",
                ...
            },
            "package": {
                "foo": {
                    "1.1.0": Definition(identifier="foo", version="1.1.0"),
                    "1.0.0": Definition(identifier="foo", version="1.0.0"),
                    "0.1.0": Definition(identifier="foo", version="0.1.0"),
                    ...
                },
                ...
            },
            "implicit-packages": [
                "bar==0.1.0",
                ...
            ]
            "registries": [
                "/path/to/registry",
                ...
            ]
        }

    :param paths: List of registry paths to recursively fetch
        :class:`definitions <wiz.definition.Definition>` from.

    :param max_depth: Limited recursion value to search for :class:`definitions
        <wiz.definition.Definition>`. Default is None, which means that all
        sub-trees will be visited.

    :param system_mapping: Mapping defining the current system to filter
        out non compatible definitions. Default is None, which means that the
        current system mapping will be :func:`queried <wiz.system.query>`.

    :return: Definition mapping.

    """
    if system_mapping is None:
        system_mapping = wiz.system.query()

    mapping = wiz.definition.fetch(
        paths, system_mapping=system_mapping, max_depth=max_depth
    )

    mapping["registries"] = paths
    return mapping


def fetch_definition(request, definition_mapping):
    """Return :class:`~wiz.definition.Definition` instance from request.

    :param request: String indicating which definition should be fetched.
        (e.g. "definition", "definition >= 1.0.0, < 2", etc.).

    :param definition_mapping: Mapping regrouping all available definitions. It
        could be fetched with :func:`fetch_definition_mapping`.

    :return: Instance of :class:`~wiz.definition.Definition`.

    :raise: :exc:`wiz.exception.RequestNotFound` is the corresponding definition
        cannot be found.

    """
    requirement = wiz.utility.get_requirement(request)
    return wiz.definition.query(
        requirement, definition_mapping[wiz.symbol.PACKAGE_REQUEST_TYPE]
    )


def fetch_package(request, definition_mapping):
    """Return best matching :class:`~wiz.package.Package` instance from request.

    :param request: String indicating which package should be fetched.
        (e.g. "package", "package[Variant] >= 1.0.0, < 2", etc.).

    :param definition_mapping: Mapping regrouping all available definitions. It
        could be fetched with :func:`fetch_definition_mapping`.

    :return: Instance of :class:`~wiz.package.Package`.

    :raise: :exc:`wiz.exception.RequestNotFound` is the corresponding definition
        cannot be found.

    .. note::

        If several packages are extracted from *request*, only the first one
        will be returned.

    """
    requirement = wiz.utility.get_requirement(request)
    packages = wiz.package.extract(
        requirement, definition_mapping[wiz.symbol.PACKAGE_REQUEST_TYPE]
    )
    return packages[0]


def fetch_package_request_from_command(command_request, definition_mapping):
    """Return package request corresponding to command request.

    Example::

        >>> mapping = {
        ...     "command": {"hiero": "nuke"},
        ...     "package": {"nuke": ...}
        ... }
        >>> fetch_package_request_from_command("hiero==10.5.*", mapping)
        nuke==10.5.*

    :param command_request: String indicating which command should be fetched.
        (e.g. "command", "command >= 1.0.0, < 2", etc.).

    :param definition_mapping: Mapping regrouping all available definitions. It
        could be fetched with :func:`fetch_definition_mapping`.

    :return: Package requests string.

    :raise: :exc:`wiz.exception.RequestNotFound` is the corresponding command
        cannot be found.

    """
    requirement = wiz.utility.get_requirement(command_request)
    request_type = wiz.symbol.COMMAND_REQUEST_TYPE

    if requirement.name not in definition_mapping[request_type]:
        raise wiz.exception.RequestNotFound(
            "No command named '{}' can be found.".format(requirement.name)
        )

    _requirement = wiz.utility.get_requirement(
        definition_mapping[request_type][requirement.name]
    )
    _requirement.specifier = requirement.specifier
    _requirement.extras = requirement.extras
    return str(_requirement)


def resolve_context(
    requests, definition_mapping=None, ignore_implicit=False,
    environ_mapping=None, maximum_combinations=None, maximum_attempts=None
):
    """Return context mapping from *requests*.

    The context should contain the resolved environment mapping, the
    resolved command mapping, and an ordered list of all serialized packages
    which constitute the resolved context.

    It should be in the form of::

        {
            "command": {
                "app": "AppExe"
                ...
            },
            "environ": {
                "KEY1": "value1",
                "KEY2": "value2",
                ...
            },
            "packages": [
                Package(identifier="test1==1.1.0", version="1.1.0"),
                Package(identifier="test2==0.3.0", version="0.3.0"),
                ...
            ],
            "registries": [
                "/path/to/registry",
                ...
            ]
        }

    :param requests: List of strings indicating the package version requested to
        build the context (e.g. ["package >= 1.0.0, < 2"])

    :param definition_mapping: Mapping regrouping all available definitions. It
        could be fetched with :func:`fetch_definition_mapping`. If no definition
        mapping is provided, a default one will be fetched from
        :func:`default registries <wiz.registry.get_defaults>`.

    :param ignore_implicit: Indicates whether implicit packages should not be
        included in context. Default is False.

    :param environ_mapping: Mapping of environment variables which would be
        augmented by the resolved environment. Default is None.

    :param maximum_combinations: Maximum number of combinations which can be
        generated from conflicting variants. Default is None, which means
        that the default value will be picked from the :ref:`configuration
        <configuration>`.

    :param maximum_attempts: Maximum number of resolution attempts before
        raising an error. Default is None, which means  that the default
        value will be picked from the :ref:`configuration <configuration>`.

    :return: Context mapping.

    :raise: :exc:`wiz.exception.GraphResolutionError` if the resolution graph
        cannot be resolved in time.

    """
    requirements = wiz.utility.get_requirements(requests)

    # Extract definition mapping from default registry paths if necessary.
    if definition_mapping is None:
        definition_mapping = wiz.fetch_definition_mapping(
            wiz.registry.get_defaults()
        )

    # Extract initial namespace counter from explicit requirements.
    namespace_counter = wiz.utility.compute_namespace_counter(
        requirements, definition_mapping[wiz.symbol.PACKAGE_REQUEST_TYPE]
    )

    # Prepend implicit requests to explicit ones if necessary.
    if not ignore_implicit:
        _requests = definition_mapping.get(wiz.symbol.IMPLICIT_PACKAGE, [])
        requirements = wiz.utility.get_requirements(_requests) + requirements

    resolver = wiz.graph.Resolver(
        definition_mapping[wiz.symbol.PACKAGE_REQUEST_TYPE],
        maximum_combinations=maximum_combinations,
        maximum_attempts=maximum_attempts,
    )
    packages = resolver.compute_packages(
        requirements, namespace_counter=namespace_counter
    )

    _environ_mapping = wiz.environ.initiate(environ_mapping)
    context = wiz.package.extract_context(
        packages, environ_mapping=_environ_mapping
    )

    context["packages"] = packages
    context["registries"] = definition_mapping["registries"]

    # Augment context environment with wiz signature
    context["environ"].update({
        "WIZ_VERSION": __version__,
        "WIZ_CONTEXT": wiz.utility.encode([
            [_package.identifier for _package in packages],
            definition_mapping["registries"]
        ])
    })
    return context


def resolve_command(elements, command_mapping):
    """Return resolved command elements from *elements* and *command_mapping*.

    Example::

        >>> resolve_command(
        ...     ["app", "--option", "value", "/path/to/script"],
        ...     {"app": "App0.1 --modeX"}
        ... )

        ["App0.1", "--modeX", "--option", "value", "/path/to/script"]

    :param elements: List of strings constituting the command line to resolve
        (e.g. ["app_exe", "--option", "value"])

    :param command_mapping: Mapping associating command aliased to real
        commands.

    :return: List of strings constituting the resolved command line.

    """
    if elements[0] in command_mapping.keys():
        elements = shlex.split(command_mapping[elements[0]]) + elements[1:]

    return elements


def discover_context():
    """Return context mapping used to resolve the current wiz environment.

    It should be in the form of::

        {
            "command": {
                "app": "AppExe"
                ...
            },
            "environ": {
                "KEY1": "value1",
                "KEY2": "value2",
                ...
            },
            "packages": [
                Package(identifier="test1==1.1.0", version="1.1.0"),
                Package(identifier="test2==0.3.0", version="0.3.0"),
                ...
            ],
            "registries": [
                "/path/to/registry",
                ...
            ]
        }

    The context should have been encoded into a :envvar:`WIZ_CONTEXT`
    environment variable that can be used to retrieve the list of registries and
    packages from which the current environment was resolved.

    .. warning::

        The context cannot be retrieved if this function is called
        outside of a resolved environment.

    :return: Context mapping.

    :raise: :exc:`~wiz.exception.RequestNotFound` if the :envvar:`WIZ_CONTEXT`
        environment variable is not found.

    """
    encoded_context = os.environ.get("WIZ_CONTEXT")
    if encoded_context is None:
        raise wiz.exception.RequestNotFound(
            "Impossible to retrieve the current context as the 'WIZ_CONTEXT' "
            "environment variable is not set. Are you sure you are currently "
            "in a resolved environment?"
        )

    package_identifiers, registries = wiz.utility.decode(encoded_context)

    # Extract and return each unique package from definition requirements.
    definition_mapping = wiz.fetch_definition_mapping(registries)
    packages = [
        wiz.fetch_package(identifier, definition_mapping)
        for identifier in package_identifiers
    ]

    _environ_mapping = wiz.environ.initiate()
    context = wiz.package.extract_context(
        packages, environ_mapping=_environ_mapping
    )

    context["packages"] = packages
    context["registries"] = registries
    return context


def load_definition(path):
    """Return :class:`~wiz.definition.Definition` instance from file *path*.

    :param path: :term:`JSON` file path which contains a definition.

    :raise: :exc:`wiz.exception.IncorrectDefinition` if the definition is
        incorrect.

    """
    return wiz.definition.load(path)


def export_definition(path, data, overwrite=False):
    """Export definition *data* as a :term:`JSON` file in directory *path*.

    :param path: Target path to save the exported definition into.

    :param data: Instance of :class:`wiz.definition.Definition` or a mapping in
        the form of::

            {
                "identifier": "foo",
                "description": "This is my package",
                "version": "0.1.0",
                "command": {
                    "app": "AppExe",
                    "appX": "AppExe --mode X"
                },
                "environ": {
                    "KEY1": "value1",
                    "KEY2": "value2"
                },
                "requirements": [
                    "package1 >=1, <2",
                    "package2"
                ]
            }

    :param overwrite: Indicate whether existing definitions in the target path
        will be overwritten. Default is False.

    :return: Path to exported definition.

    :raise: :exc:`wiz.exception.IncorrectDefinition` if *data* is a mapping that
        cannot create a valid instance of :class:`wiz.definition.Definition`.

    :raise: :exc:`wiz.exception.FileExists` if definition already exists in
        *path* and overwrite is False.

    :raise: :exc:`OSError` if the definition can not be exported in *path*.

    .. warning::

        Ensure that the *data* :ref:`identifier <definition/identifier>`,
        :ref:`namespace <definition/namespace>`, :ref:`version
        <definition/version>` and :ref:`system requirement <definition/system>`
        are unique in the registry.

        Each :ref:`command <definition/command>` must also be unique in the
        registry.

    """
    return wiz.definition.export(path, data, overwrite=overwrite)


def export_script(
    path, script_type, identifier, environ, command=None, packages=None,
):
    """Export environment as :term:`Bash` or :term:`C-Shell` script in *path*.

    :param path: Target path to save the exported script into.

    :param script_type: Should be either :term:`tcsh <C-Shell>` or
        :term:`bash <Bash>`.

    :param identifier: File name of the exported script.

    :param environ: Mapping of all environment variables that will be set by the
        exported definition. It should be in the form of::

            {
                "KEY1": "value1",
                "KEY2": "value2",
            }

    :param command: Define a command to run within the exported wrapper. Default
        is None.

    :param packages: Indicate a list of :class:`wiz.package.Package` instances
        which helped creating the context.

    :return: Path to the exported script.

    :raise: :exc:`ValueError` if the *script_type* is incorrect.

    :raise: :exc:`ValueError` if *environ* mapping is empty.

    :raise: :exc:`OSError` if the wrapper can not be exported in *path*.

    """
    file_path = os.path.join(os.path.abspath(path), identifier)

    if script_type == "bash":
        content = "#!/bin/bash\n"
    elif script_type == "tcsh":
        content = "#!/bin/tcsh -f\n"
    else:
        raise ValueError("'{}' is not a valid script type.".format(script_type))

    # Indicate information about the generation process.
    if packages is not None:
        content += "#\n# Generated by wiz with the following environments:\n"
        for _package in packages:
            content += "# - {}\n".format(_package.identifier)
        content += "#\n"

    if len(environ.keys()) == 0:
        raise ValueError("The environment mapping should not be empty.")

    for key, value in environ.items():
        # Do not override the PATH environment variable to prevent
        # error when executing the script.
        if key == "PATH":
            value += ":${PATH}"

        if script_type == "bash":
            content += "export {0}=\"{1}\"\n".format(key, value)
        else:
            content += "setenv {0} \"{1}\"\n".format(key, value)

    if command is not None:
        if script_type == "bash":
            content += command + " $@\n"
        else:
            content += command + " $argv:q\n"

    wiz.filesystem.export(file_path, content)
    return file_path


def validate_definition(definition, definition_mapping=None):
    """Return validation mapping for *definition*.

    :param definition: Instance of :class:`wiz.definition.Definition`.

    :param definition_mapping: Mapping regrouping all available definitions. It
        could be fetched with :func:`fetch_definition_mapping`. If no definition
        mapping is provided, a default one will be fetched from
        :func:`default registries <wiz.registry.get_defaults>`.

    :return: Mapping in the form of
        ::

            {
                "errors": [],
                "warnings": []
            }

    """
    if definition_mapping is None:
        definition_mapping = wiz.fetch_definition_mapping(
            wiz.registry.get_defaults()
        )

    # Record current logging handlers to set it back once validation is done.
    _handlers = wiz.logging.root.handlers.copy()

    # Configure logger so that errors and warnings are recorded and not
    # directly displayed.
    log_error, log_warning = wiz.logging.configure_for_debug()

    mapping = _fetch_validation_mapping(
        log_error, log_warning, definition, definition_mapping
    )

    # Reset logging handlers.
    wiz.logging.root.handlers = _handlers

    return mapping


def _fetch_validation_mapping(
    log_error, log_warning, definition, definition_mapping
):
    """Fetch errors and warnings from definition for *definition*.

    :param log_error: instances of :class:`io.StringIO` such as
        the instance returned by :func:`wiz.logging.configure_for_debug`. It
        will receive possible error logged during the context resolution
        process.

    :param log_warning: instances of :class:`io.StringIO` such as
        the instance returned by :func:`wiz.logging.configure_for_debug`. It
        will receive possible warning logged during the context resolution
        process.

    :param definition: instance of :class:`wiz.definition.Definition`.

    :param definition_mapping: Mapping regrouping all available definitions. It
        could be fetched with :func:`fetch_definition_mapping`.

    :return: Mapping in the form of
        ::

            {
                "errors": [],
                "warnings": []
            }

    .. warning::

        *log_error* and *log_warning* variables will be :meth:`closed
        <io.StringIO.close>` and should not be re-used.

    """
    mapping = {"errors": [], "warnings": []}

    try:
        _context = wiz.resolve_context(
            [definition.qualified_version_identifier],
            definition_mapping,
            ignore_implicit=True,
        )

        error = log_error.getvalue()
        warning = log_warning.getvalue()

        if len(error) > 0:
            mapping["errors"].append(error)

        if len(warning) > 0:
            mapping["warnings"].append(warning)

        log_error.close()
        log_warning.close()

    except wiz.exception.WizError as error:
        mapping["errors"].append("critical: {}".format(str(error)))

    else:
        for key, value in _context.get("environ", {}).items():
            unresolved_variables = []
            for variable_tuple in wiz.environ.ENV_PATTERN.findall(value):
                unresolved_variables.append("".join(variable_tuple))

            if len(unresolved_variables) > 0:
                warning = (
                    "warning: the '{}' environment variable contains "
                    "unresolved elements: {}".format(
                        key, ", ".join(unresolved_variables)
                    )
                )
                mapping["warnings"].append(warning)

    return mapping



# :coding: utf-8

import pytest
import six

import wiz.validator
import wiz.exception


def test_validate_definition_minimal():
    """Validate minimal definition data mapping."""
    wiz.validator.validate_definition({"identifier": "foo"})


def test_validate_definition_with_version():
    """Validate definition data mapping with 'version' keyword."""
    wiz.validator.validate_definition({
        "identifier": "foo",
        "version": "0.1.0"
    })


def test_validate_definition_with_description():
    """Validate definition data mapping with 'description' keyword."""
    wiz.validator.validate_definition({
        "identifier": "foo",
        "description": "This is a definition",
    })


def test_validate_definition_with_disabled():
    """Validate definition data mapping with 'disabled' keyword."""
    wiz.validator.validate_definition({
        "identifier": "foo",
        "disabled": True,
    })


def test_validate_definition_with_auto_use():
    """Validate definition data mapping with 'auto-use' keyword."""
    wiz.validator.validate_definition({
        "identifier": "foo",
        "auto-use": True,
    })


def test_validate_definition_with_system():
    """Validate definition data mapping with 'system' keyword."""
    wiz.validator.validate_definition({
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition",
        "system": {
            "platform": "linux",
            "os": "el >=7, <8",
            "arch": "x86_64"
        },
    })


def test_validate_definition_with_environ():
    """Validate definition data mapping with 'environ' keyword."""
    wiz.validator.validate_definition({
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition",
        "environ": {
            "KEY1": "VALUE1",
            "KEY2": "VALUE2",
            "KEY3": "PATH1:PATH2:PATH3"
        },
    })


def test_validate_definition_with_command():
    """Validate definition data mapping with 'command' keyword."""
    wiz.validator.validate_definition({
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition",
        "command": {
            "app": "App0.1",
            "appX": "App0.1 --option value"
        },
    })


def test_validate_definition_with_requirements():
    """Validate definition data mapping with 'requirements' keyword."""
    wiz.validator.validate_definition({
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition",
        "requirements": [
            "envA >= 1.0.0",
            "envB >= 3.4.2, < 4",
            "envC"
        ],
    })


def test_validate_definition_with_conditions():
    """Validate definition data mapping with 'conditions' keyword."""
    wiz.validator.validate_definition({
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition",
        "conditions": [
            "envA >= 1.0.0",
            "envB >= 3.4.2, < 4",
            "envC"
        ],
    })


def test_validate_definition_with_variants():
    """Validate definition data mapping with 'variants' keyword."""
    wiz.validator.validate_definition({
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition",
        "variants": [
            {
                "identifier": "1.0",
                "environ": {
                    "VERSION": "1.0"
                },
                "requirements": [
                    "envA >= 1.0, < 2"
                ]
            },
            {
                "identifier": "2.0",
                "environ": {
                    "VERSION": "2.0"
                },
                "command": {
                    "app": "App2.0",
                },
                "requirements": [
                    "envA >= 2.0, < 3"
                ]
            },
            {
                "identifier": "XXX",
                "command": {
                    "app": "AppXXX",
                },
            }
        ],
    })


@pytest.mark.parametrize("value, message", [
    (42, "Data has incorrect type."),
    ({"test": "foo"}, "Data contains invalid keywords: test"),
    ({}, "'identifier' is required."),
    ({"identifier": 0}, "'identifier' has incorrect type."),
    ({"identifier": "foo", "version": 0}, "'version' has incorrect type."),
    (
        {"identifier": "foo", "description": 0},
        "'description' has incorrect type."
    ),
    ({"identifier": "foo", "disabled": 0}, "'disabled' has incorrect type."),
    ({"identifier": "foo", "system": 0}, "'system' has incorrect type."),
    ({"identifier": "foo", "environ": 0}, "'environ' has incorrect type."),
    ({"identifier": "foo", "command": 0}, "'command' has incorrect type."),
    (
        {"identifier": "foo", "requirements": 0},
        "'requirements' has incorrect type."
    ),
    (
        {"identifier": "foo", "conditions": 0},
        "'conditions' has incorrect type."
    ),
    ({"identifier": "foo", "variants": 0}, "'variants' has incorrect type."),
], ids=[
    "incorrect-type",
    "invalid-keywords",
    "identifier-missing",
    "identifier-incorrect",
    "version-incorrect",
    "description-incorrect",
    "disabled-incorrect",
    "system-incorrect",
    "environ-incorrect",
    "command-incorrect",
    "requirements-incorrect",
    "conditions-incorrect",
    "variants-incorrect",
])
def test_validate_definition_failed(value, message):
    """Raise error when data is incorrect."""
    with pytest.raises(wiz.exception.DefinitionError) as error:
        wiz.validator.validate_definition(value)

    assert message in str(error)


def test_validate_identifier_keyword():
    """Validate 'identifier' keyword within data."""
    wiz.validator.validate_identifier_keyword({"identifier": "foo"})


@pytest.mark.parametrize("value, options, message", [
    ({}, {}, "'identifier' is required."),
    ({"identifier": 42}, {}, "'identifier' has incorrect type."),
    ({}, {"variant_index": 1}, "'variants/1/identifier' is required."),
], ids=[
    "required",
    "incorrect-type",
    "variant-index",
])
def test_validate_identifier_keyword_failed(value, options, message):
    """Raise error when 'identifier' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_identifier_keyword(value, **options)

    assert message in str(error)


def test_validate_version_keyword():
    """Validate 'version' keyword within data."""
    wiz.validator.validate_version_keyword({})
    wiz.validator.validate_version_keyword({"version": "0.1.0"})


@pytest.mark.parametrize("value, message", [
    ({"version": 42}, "'version' has incorrect type."),
    ({"version": "_"}, "Invalid version: '_'"),
    ({"version": "abc"}, "Invalid version: 'abc'"),
    ({"version": "test0.1.0"}, "Invalid version: 'test0.1.0'"),
    ({"version": "0.1.*"}, "Invalid version: '0.1.*'"),
    ({"version": "0.1."}, "Invalid version: '0.1.'"),
    ({"version": "#@;"}, "Invalid version: '#@;'"),
], ids=[
    "incorrect-type",
    "incorrect-version-1",
    "incorrect-version-2",
    "incorrect-version-3",
    "incorrect-version-4",
    "incorrect-version-5",
    "incorrect-version-6",
])
def test_validate_version_keyword_failed(value, message):
    """Raise error when 'version' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_version_keyword(value)

    assert message in str(error)


def test_validate_namespace_keyword():
    """Validate 'namespace' keyword within data."""
    wiz.validator.validate_namespace_keyword({})
    wiz.validator.validate_namespace_keyword({"namespace": "foo"})


def test_validate_namespace_keyword_failed():
    """Raise error when 'namespace' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_namespace_keyword({"namespace": True})

    assert "'namespace' has incorrect type." in str(error)


def test_validate_description_keyword():
    """Validate 'description' keyword within data."""
    wiz.validator.validate_description_keyword({})
    wiz.validator.validate_description_keyword({"description": "test"})


def test_validate_description_keyword_failed():
    """Raise error when 'description' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_description_keyword({"description": True})

    assert "'description' has incorrect type." in str(error)


def test_validate_auto_use_keyword():
    """Validate 'auto-use' keyword within data."""
    wiz.validator.validate_auto_use_keyword({})
    wiz.validator.validate_auto_use_keyword({"auto-use": True})


def test_validate_auto_use_keyword_failed():
    """Raise error when 'auto-use' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_auto_use_keyword({"auto-use": "foo"})

    assert "'auto-use' has incorrect type." in str(error)


def test_validate_disabled_keyword():
    """Validate 'disabled' keyword within data."""
    wiz.validator.validate_disabled_keyword({})
    wiz.validator.validate_disabled_keyword({"disabled": True})


def test_validate_disabled_keyword_failed():
    """Raise error when 'disabled' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_disabled_keyword({"disabled": "foo"})

    assert "'disabled' has incorrect type." in str(error)


def test_validate_install_root_keyword():
    """Validate 'install-root' keyword within data."""
    wiz.validator.validate_install_root_keyword({})
    wiz.validator.validate_install_root_keyword({
        "install-root": "/path/to/install/root"
    })


def test_validate_install_root_keyword_failed():
    """Raise error when 'install-root' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_install_root_keyword({"install-root": False})

    assert "'install-root' has incorrect type." in str(error)


def test_validate_install_location_keyword():
    """Validate 'install-location' keyword within data."""
    wiz.validator.validate_install_location_keyword({})
    wiz.validator.validate_install_location_keyword({
        "install-location": "/path/to/install/location"
    })


def test_validate_install_location_keyword_failed():
    """Raise error when 'install-location' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_install_location_keyword({"install-location": 0})

    assert "'install-location' has incorrect type." in str(error)


def test_validate_system_keyword():
    """Validate 'system' keyword within data."""
    wiz.validator.validate_system_keyword({})
    wiz.validator.validate_system_keyword({"system": {"platform": "linux"}})
    wiz.validator.validate_system_keyword({"system": {"os": "el >= 7.4"}})
    wiz.validator.validate_system_keyword({"system": {"arch": "x86_64"}})


@pytest.mark.parametrize("value, message", [
    ({"system": {}}, "'system' should not be empty."),
    ({"system": 42}, "'system' has incorrect type."),
    ({"system": {"test": "foo"}}, "'system' contains invalid keywords: test"),
    ({"system": {"platform": 42}}, "system/platform' has incorrect type."),
    ({"system": {"os": 42}}, "system/os' has incorrect type."),
    ({"system": {"arch": 42}}, "system/arch' has incorrect type."),
], ids=[
    "empty",
    "incorrect-type",
    "invalid-keywords",
    "incorrect-platform",
    "incorrect-os",
    "incorrect-arch",
])
def test_validate_system_keyword_failed(value, message):
    """Raise error when 'identifier' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_system_keyword(value)

    assert message in str(error)


def test_validate_command_keyword():
    """Validate 'command' keyword within data."""
    wiz.validator.validate_command_keyword({})
    wiz.validator.validate_command_keyword({"command": {"foo": "FooExe"}})


@pytest.mark.parametrize("value, options, message", [
    ({"command": {}}, {}, "'command' should not be empty."),
    ({"command": 42}, {}, "'command' has incorrect type."),
    (
        {"command": 42}, {"variant_index": 1},
        "'variants/1/command' has incorrect type."
    ),
], ids=[
    "empty",
    "incorrect-type",
    "variant-index",
])
def test_validate_command_keyword_failed(value, options, message):
    """Raise error when 'command' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_command_keyword(value, **options)

    assert message in str(error)


def test_validate_environ_keyword():
    """Validate 'environ' keyword within data."""
    wiz.validator.validate_environ_keyword({})
    wiz.validator.validate_environ_keyword({"environ": {"key": "value"}})


@pytest.mark.parametrize("value, options, message", [
    ({"environ": {}}, {}, "'environ' should not be empty."),
    ({"environ": 42}, {}, "'environ' has incorrect type."),
    (
        {"environ": 42}, {"variant_index": 1},
        "'variants/1/environ' has incorrect type."
    ),
], ids=[
    "empty",
    "incorrect-type",
    "variant-index",
])
def test_validate_environ_keyword_failed(value, options, message):
    """Raise error when 'environ' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_environ_keyword(value, **options)

    assert message in str(error)


def test_validate_requirements_keyword():
    """Validate 'requirements' keyword within data."""
    wiz.validator.validate_requirements_keyword({})
    wiz.validator.validate_requirements_keyword({"requirements": ["foo"]})


@pytest.mark.parametrize("value, options, message", [
    ({"requirements": []}, {}, "'requirements' should not be empty."),
    ({"requirements": 42}, {}, "'requirements' has incorrect type."),
    (
        {"requirements": 42}, {"variant_index": 1},
        "'variants/1/requirements' has incorrect type."
    ),
], ids=[
    "empty",
    "incorrect-type",
    "variant-index",
])
def test_validate_requirements_keyword_failed(value, options, message):
    """Raise error when 'requirements' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_requirements_keyword(value, **options)

    assert message in str(error)


def test_validate_conditions_keyword():
    """Validate 'conditions' keyword within data."""
    wiz.validator.validate_conditions_keyword({})
    wiz.validator.validate_conditions_keyword({"conditions": ["foo"]})


@pytest.mark.parametrize("value, message", [
    ({"conditions": []}, "'conditions' should not be empty."),
    ({"conditions": 42}, "'conditions' has incorrect type."),
], ids=[
    "empty",
    "incorrect-type",
])
def test_validate_conditions_keyword_failed(value, message):
    """Raise error when 'conditions' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_conditions_keyword(value)

    assert message in str(error)


def test_validate_variants_keyword():
    """Validate 'variants' keyword within data."""
    wiz.validator.validate_variants_keyword({})
    wiz.validator.validate_variants_keyword(
        {
            "variants": [
                {
                    "identifier": "foo",
                    "install-location": "/path/to/install/location",
                    "command": {"foo": "FooExe"},
                    "environ": {"key": "value"},
                    "requirements": ["foo >= 0.1.0"]
                }
            ]
        }
    )


@pytest.mark.parametrize("value, message", [
    ({"variants": []}, "'variants' should not be empty."),
    ({"variants": 42}, "'variants' has incorrect type."),
    ({"variants": [42]}, "'variants/0' has incorrect type."),
    ({"variants": [{"A": "foo"}]}, "'variants/0' contains invalid keywords: A"),
    ({"variants": [{}]}, "'variants/0/identifier' is required."),
    (
        {"variants": [{"identifier": 42}]},
        "'variants/0/identifier' has incorrect type."
    ),
    (
        {"variants": [{"identifier": "foo", "install-location": 42}]},
        "'variants/0/install-location' has incorrect type."
    ),
    (
        {"variants": [{"identifier": "foo", "command": 42}]},
        "'variants/0/command' has incorrect type."
    ),
    (
        {"variants": [{"identifier": "foo", "environ": 42}]},
        "'variants/0/environ' has incorrect type."
    ),
    (
        {"variants": [{"identifier": "foo", "requirements": 42}]},
        "'variants/0/requirements' has incorrect type."
    ),
], ids=[
    "empty",
    "incorrect-type",
    "variant-incorrect-type",
    "variant-invalid-keywords",
    "variant-identifier-missing",
    "variant-identifier-incorrect",
    "variant-install-location-incorrect",
    "variant-command-incorrect",
    "variant-environ-incorrect",
    "variant-requirements-incorrect",
])
def test_validate_variants_keyword_failed(value, message):
    """Raise error when 'variants' keyword is incorrect."""
    with pytest.raises(ValueError) as error:
        wiz.validator.validate_variants_keyword(value)

    assert message in str(error)


def test_validate_keywords():
    """Ensure that no invalid keywords are in data mapping."""
    keywords = {"A", "B"}

    wiz.validator.validate_keywords({}, keywords)
    wiz.validator.validate_keywords({"A": "foo"}, keywords)
    wiz.validator.validate_keywords({"A": "foo", "B": "bar"}, keywords)

    with pytest.raises(ValueError) as error:
        wiz.validator.validate_keywords({"A": "foo", "C": "bar"}, keywords)

    assert "Data contains invalid keywords: C" in str(error)

    with pytest.raises(ValueError) as error:
        wiz.validator.validate_keywords({"C": "bar"}, keywords, label="'test'")

    assert "'test' contains invalid keywords: C" in str(error)


def test_validate_required():
    """Ensure that data exists."""
    wiz.validator.validate_required("")
    wiz.validator.validate_required("foo")
    wiz.validator.validate_required(0)
    wiz.validator.validate_required(42)
    wiz.validator.validate_required(False)
    wiz.validator.validate_required(True)
    wiz.validator.validate_required([])
    wiz.validator.validate_required([1, 2, 3])
    wiz.validator.validate_required({})
    wiz.validator.validate_required({"A": "foo"})
    wiz.validator.validate_required({1, 2, 3})

    with pytest.raises(ValueError) as error:
        wiz.validator.validate_required(None)

    assert "Data is required." in str(error)

    with pytest.raises(ValueError) as error:
        wiz.validator.validate_required(None, label="'test'")

    assert "'test' is required." in str(error)


def test_validate_type():
    """Ensure that data has correct type."""
    wiz.validator.validate_type("foo", six.string_types)
    wiz.validator.validate_type("0.1.0", six.string_types)
    wiz.validator.validate_type(42, int)
    wiz.validator.validate_type([1, 2, 3], list)
    wiz.validator.validate_type({"A": "foo"}, dict)
    wiz.validator.validate_type(True, bool)

    with pytest.raises(ValueError) as error:
        wiz.validator.validate_type(42, six.string_types)

    assert "Data has incorrect type." in str(error)

    with pytest.raises(ValueError) as error:
        wiz.validator.validate_type(42, six.string_types, label="'test'")

    assert "'test' has incorrect type." in str(error)


def test_validate_not_empty():
    """Ensure that data container is not empty."""
    wiz.validator.validate_not_empty({"A": "foo"})
    wiz.validator.validate_not_empty([1, 2, 3])

    with pytest.raises(ValueError) as error:
        wiz.validator.validate_not_empty({})

    assert "Data should not be empty." in str(error)

    with pytest.raises(ValueError) as error:
        wiz.validator.validate_not_empty([], label="'test'")

    assert "'test' should not be empty." in str(error)



# :coding: utf-8

import os
import copy
import json
import collections

import ujson

import wiz.exception
import wiz.filesystem
import wiz.history
import wiz.logging
import wiz.package
import wiz.symbol
import wiz.system
import wiz.utility
import wiz.validator


def fetch(paths, system_mapping=None, max_depth=None):
    """Return mapping from all definitions available under *paths*.

    A definition mapping should be in the form of::

        {
            "command": {
                "fooExe": "foo",
                ...
            },
            "package": {
                "__namespace__": {
                    "bar": {"test"}
                },
                "foo": {
                    "1.1.0": <Definition(identifier="foo", version="1.1.0")>,
                    "1.0.0": <Definition(identifier="foo", version="1.0.0")>,
                    "0.1.0": <Definition(identifier="foo", version="0.1.0")>,
                    ...
                },
                "test::bar": {
                    "0.1.0": <Definition(identifier="bar", version="0.1.0")>,
                    ...
                },
                ...
            },
            "implicit-packages": [
                "bar==0.1.0",
                ...
            ]
        }

    :param paths: List of registry paths to recursively fetch
        :class:`definitions <Definition>` from.

    :param system_mapping: Mapping defining the current system to filter
        out non compatible definitions. Default is None, which means that the
        current system mapping will be :func:`queried <wiz.system.query>`.

    :param max_depth: Limited recursion value to search for :class:`definitions
        <Definition>`. Default is None, which means that all  sub-trees will be
        visited.

    :return: Definition mapping.

    """
    mapping = {
        wiz.symbol.PACKAGE_REQUEST_TYPE: {},
        wiz.symbol.COMMAND_REQUEST_TYPE: {},
    }

    # Record definitions which should be implicitly used.
    implicit_identifiers = []
    implicit_package_mapping = {}

    for definition in discover(
        paths, system_mapping=system_mapping, max_depth=max_depth
    ):
        _add_to_mapping(definition, mapping[wiz.symbol.PACKAGE_REQUEST_TYPE])

        # Record commands from definition.
        for command in definition.command.keys():
            mapping[wiz.symbol.COMMAND_REQUEST_TYPE][command] = (
                definition.qualified_identifier
            )

        # Record package identifiers which should be used implicitly in context.
        if definition.auto_use:
            implicit_identifiers.append(definition.qualified_identifier)
            _add_to_mapping(definition, implicit_package_mapping)

    # Extract implicit package requests.
    mapping[wiz.symbol.IMPLICIT_PACKAGE] = _extract_implicit_requests(
        implicit_identifiers, implicit_package_mapping
    )

    wiz.history.record_action(
        wiz.symbol.DEFINITIONS_COLLECTION_ACTION,
        registries=paths, max_depth=max_depth, definition_mapping=mapping
    )

    return mapping


def _add_to_mapping(definition, mapping):
    """Mutate package *mapping* to add *definition*

    The mutated mapping should be in the form of::

        {
            "foo": {
                "1.1.0": <Definition(identifier="foo", version="1.1.0")>,
                "1.0.0": <Definition(identifier="foo", version="1.0.0")>,
                "0.1.0": <Definition(identifier="foo", version="0.1.0")>,
                ...
            },
            ...
        }

    :param definition: Instance of :class:`Definition`.

    :param mapping: Mapping to mutate.

    """
    identifier = definition.identifier
    if definition.namespace is not None:
        mapping.setdefault("__namespace__", {})
        mapping["__namespace__"].setdefault(identifier, set())
        mapping["__namespace__"][identifier].add(definition.namespace)

    qualified_identifier = definition.qualified_identifier
    version = str(definition.version or wiz.symbol.UNSET_VALUE)

    mapping.setdefault(qualified_identifier, {})
    mapping[qualified_identifier].setdefault(version, {})
    mapping[qualified_identifier][version] = definition


def _extract_implicit_requests(identifiers, mapping):
    """Extract requests from definition *identifiers* and package *mapping*.

    Package requests are returned in inverse order of discovery to give priority
    to the latest discovered

    :param identifiers: List of definition identifiers sorted in order of
        discovery.

    :param mapping: Mapping regrouping all implicit package definitions.
        It should be in the form of::

            {
                "__namespace__": {
                    "bar": {"test"}
                },
                "foo": {
                    "1.1.0": <Definition(identifier="foo", version="1.1.0")>,
                    "1.0.0": <Definition(identifier="foo", version="1.0.0")>,
                    "0.1.0": <Definition(identifier="foo", version="0.1.0")>,
                    ...
                },
                "test::bar": {
                    "0.1.0": <Definition(identifier="bar", version="0.1.0")>,
                    ...
                },
                ...
        }

    :return: List of request strings.

    """
    requests = []

    for identifier in sorted(
        (_id for _id in mapping.keys() if _id != "__namespace__"),
        key=lambda _id: identifiers.index(_id), reverse=True
    ):
        requirement = wiz.utility.get_requirement(identifier)
        definition = query(requirement, mapping)
        requests.append(definition.qualified_version_identifier)

    return requests


def query(requirement, definition_mapping, namespace_counter=None):
    """Return best matching definition version from *requirement*.

    :param requirement: Instance of :class:`packaging.requirements.Requirement`.

    :param definition_mapping: Mapping regrouping all available definitions
        associated with their unique identifier.

    :param namespace_counter: instance of :class:`collections.Counter`
        which indicates occurrence of namespaces used as hints for package
        identification. Default is None.

    :return: Instance of :class:`Definition`.

    :raise: :exc:`wiz.exception.RequestNotFound` if the requirement can not
        be resolved.

    """
    identifier = requirement.name

    variant_identifier = None

    # Extract variant if necessary.
    if len(requirement.extras) > 0:
        variant_identifier = next(iter(requirement.extras))

    # Extend identifier with namespace if necessary.
    if wiz.symbol.NAMESPACE_SEPARATOR not in identifier:
        identifier = _guess_qualified_identifier(
            identifier, definition_mapping, namespace_counter=namespace_counter
        )

    # If identifier starts with namespace separator, that means the identifier
    # without namespace is required.
    if identifier.startswith(wiz.symbol.NAMESPACE_SEPARATOR):
        identifier = identifier[2:]

    if identifier not in definition_mapping:
        raise wiz.exception.RequestNotFound(
            "The requirement '{}' could not be resolved.".format(requirement)
        )

    definition = None

    # Extract all versions from definitions.
    versions = [
        _definition.version or wiz.symbol.UNSET_VALUE
        for _definition in definition_mapping[identifier].values()
    ]

    if wiz.symbol.UNSET_VALUE in versions and len(versions) > 1:
        raise wiz.exception.RequestNotFound(
            "Impossible to retrieve the best matching definition for "
            "'{}' as non-versioned and versioned definitions have "
            "been fetched.".format(identifier)
        )

    # Sort the definition versions so that the highest one is first.
    versions = sorted(versions, reverse=True)

    # Get the best matching definition by sorting versions so that the highest
    # one is first.
    for version in sorted(versions, reverse=True):
        _definition = definition_mapping[identifier][str(version)]

        # Skip if the variant identifier required is not found in definition.
        if variant_identifier and not any(
            variant_identifier == variant.identifier
            for variant in _definition.variants
        ):
            continue

        if (
            _definition.version is None
            or _definition.version in requirement.specifier
        ):
            definition = _definition
            break

    if definition is None:
        raise wiz.exception.RequestNotFound(
            "The requirement '{}' could not be resolved.".format(requirement)
        )

    return definition


def _guess_qualified_identifier(
    identifier, definition_mapping, namespace_counter=None
):
    """Return qualified identifier with default namespace if possible.

    Rules are as follow:

    * If definition does not have any namespaces, return identifier;
    * If definition has one namespace, return identifier with namespace;
    * If definition has one namespace and also exists without identifier,
      return the one without a namespace;
    * If definition has several namespaces available, use the
      *namespace_counter* to filter out namespaces which don't have the maximum
      occurrence number. If only one namespace remains, use this one;
    * If definition still has several namespaces available after checking
      occurrences, if one of these namespaces is equal to the identifier (e.g.
      "maya::maya"), return that one.
    * If definition still has several namespaces after exhausting all other
      options, raise :exc:`wiz.exception.RequestNotFound`.

    :param identifier: Unique identifier of a definition.

    :param definition_mapping: Mapping regrouping all available definitions
        associated with their unique identifier.

    :param namespace_counter: instance of :class:`collections.Counter`
        which indicates occurrence of namespaces used as hints for package
        identification. Default is None.

    :return: Qualified identifier (e.g. "namespace::foo")

    :raise: :exc:`wiz.exception.RequestNotFound` if the default namespace can
        not be guessed.

    """
    namespace_mapping = definition_mapping.get("__namespace__", {})

    _namespaces = list(namespace_mapping.get(identifier, []))

    # If no namespace are found, just return identifier unchanged.
    if len(_namespaces) == 0:
        return identifier

    max_occurrence = 0

    # Fetch number of occurrence of the namespace for counter if available.
    if namespace_counter is not None:
        max_occurrence = max([namespace_counter[name] for name in _namespaces])

    # If more than one namespace is available, attempt to use counter to only
    # keep those which are used the most.
    if len(_namespaces) > 1 and max_occurrence > 0:
        _namespaces = [
            namespace for namespace in _namespaces
            if namespace_counter[namespace] == max_occurrence
        ]

    # If more than one namespace is available and one namespace is identical to
    # the definition identifier (e.g. "maya::maya"), it will be selected by
    # default.
    if len(_namespaces) > 1 and any(name == identifier for name in _namespaces):
        _namespaces = [identifier]

    # If more than one namespace is available or if we didn't need to shrink the
    # namespace list from occurrences and definition exists without a namespace,
    # it will be selected by default.
    if (
        (len(_namespaces) > 1 or max_occurrence <= 1)
        and identifier in definition_mapping
    ):
        return identifier

    if len(_namespaces) == 1:
        return _namespaces.pop() + wiz.symbol.NAMESPACE_SEPARATOR + identifier

    raise wiz.exception.RequestNotFound(
        "Cannot guess default namespace for '{definition}' "
        "[available: {namespaces}].".format(
            definition=identifier,
            namespaces=", ".join(sorted(_namespaces))
        )
    )


def export(path, data, overwrite=False):
    """Export *definition* as a :term:`JSON` file to *path*.

    :param path: Target path to save the exported definition into.

    :param data: Instance of :class:`wiz.definition.Definition` or a mapping in
        the form of::

            {
                "identifier": "foo",
                "description": "This is my package",
                "version": "0.1.0",
                "command": {
                    "app": "AppExe",
                    "appX": "AppExe --mode X"
                },
                "environ": {
                    "KEY1": "value1",
                    "KEY2": "value2"
                },
                "requirements": [
                    "package1 >=1, <2",
                    "package2"
                ]
            }

    :param overwrite: Indicate whether existing definitions in the target path
        will be overwritten. Default is False.

    :return: Path to exported definition.

    :raise: :exc:`wiz.exception.IncorrectDefinition` if *data* is a mapping that
        cannot create a valid instance of :class:`wiz.definition.Definition`.

    :raise: :exc:`wiz.exception.FileExists` if definition already exists in
        *path* and overwrite is False.

    :raise: :exc:`OSError` if the definition can not be exported in *path*.

    .. warning::

        Ensure that the *data* :ref:`identifier <definition/identifier>`,
        :ref:`namespace <definition/namespace>`, :ref:`version
        <definition/version>` and :ref:`system requirement <definition/system>`
        are unique in the registry.

        Each :ref:`command <definition/command>` must also be unique in the
        registry.

    """
    if not isinstance(data, Definition):
        definition = wiz.definition.Definition(data)
    else:
        definition = data

    file_name = wiz.utility.compute_file_name(definition)
    file_path = os.path.join(os.path.abspath(path), file_name)
    wiz.filesystem.export(file_path, definition.encode(), overwrite=overwrite)
    return file_path


def discover(paths, system_mapping=None, max_depth=None):
    """Discover and yield all definitions found under *paths*.

    :param paths: List of registry paths to recursively fetch
        :class:`definitions <Definition>` from.

    :param system_mapping: Mapping of the current system which will filter out
        non compatible definitions. The mapping should have been retrieved via
        :func:`wiz.system.query`.

    :param max_depth: Limited recursion value to search for :class:`definitions
        <Definition>`. Default is None, which means that all  sub-trees will be
        visited.

    :return: Generator which yield all :class:`definitions <Definition>`.

    """
    logger = wiz.logging.Logger(__name__ + ".discover")

    for path in paths:

        # Ignore empty paths that could resolve to current directory.
        path = path.strip()
        if not path:
            logger.debug("Skipping empty path.")
            continue

        path = os.path.abspath(path)
        logger.debug("Searching under {!r} for definition files.".format(path))

        initial_depth = path.rstrip(os.sep).count(os.sep)
        for base, _, filenames in os.walk(path):
            depth = base.count(os.sep)
            if max_depth is not None and (depth - initial_depth) > max_depth:
                continue

            for filename in filenames:
                _, extension = os.path.splitext(filename)
                if extension != ".json":
                    continue

                _path = os.path.join(base, filename)

                # Load and validate the definition.
                try:
                    definition = load(_path, registry_path=path)

                except (
                    IOError, ValueError, TypeError,
                    wiz.exception.WizError
                ):
                    logger.warning(
                        "Error occurred trying to load definition from {!r}"
                        .format(_path),
                    )
                    logger.debug_traceback()
                    continue

                # Skip definition if an incompatible system if set.
                if (
                    system_mapping is not None and
                    not wiz.system.validate(definition, system_mapping)
                ):
                    continue

                # Skip definition if "disabled" keyword is set to True.
                if definition.disabled:
                    _id = definition.qualified_version_identifier
                    logger.warning("Definition '{}' is disabled".format(_id))
                    continue

                yield definition


def load(path, mapping=None, registry_path=None):
    """Load and return a definition from *path*.

    :param path: :term:`JSON` file path which contains a definition.

    :param mapping: Mapping which will augment the data leading to the creation
        of the definition. Default is None.

    :param registry_path: Path to the registry which contains the definition.
        Default is None.

    :return: Instance of :class:`Definition`.

    :raise: :exc:`wiz.exception.IncorrectDefinition` if the definition is
        incorrect.

    """
    if mapping is None:
        mapping = {}

    with open(path, "r") as stream:
        definition_data = ujson.load(stream)
        definition_data.update(mapping)

        return Definition(
            definition_data,
            path=path,
            registry_path=registry_path,
            copy_data=False
        )


class Definition(object):
    """Definition object."""

    def __init__(
        self, data, path=None, registry_path=None, copy_data=True
    ):
        """Initialize definition from input *data* mapping.

        :param data: Data definition mapping.

        :param path: Path to the definition :term:`JSON` file used to create the
            definition if available. Default is None.

        :param registry_path: Path to the registry from which the definition
            where fetched if available. Default is None.

        :param copy_data: Indicate whether input *data* will be copied to
            prevent mutating it. Default is True.

        :raise: :exc:`wiz.exception.IncorrectDefinition` if the *data* mapping
            is incorrect.

        .. warning::

            "requirements" and "conditions" values will not get validated when
            constructing the instance for performance reason. Therefore,
            accessing these values could raise an error when data is incorrect::

                >>> definition = Definition({
                ...     "identifier": "foo",
                ...     "requirements": ["!!!"],
                ... })
                >>> print(definition.requirements)

                InvalidRequirement: The requirement '!!!' is incorrect

        .. seealso:: :ref:`definition`

        """
        wiz.validator.validate_definition(data)

        # Ensure that input data is not mutated if requested.
        if copy_data:
            data = copy.deepcopy(data)

        self._data = data
        self._path = path
        self._registry_path = registry_path

        # Store values that needs to be constructed.
        self._cache = {}

    def __repr__(self):
        """Representing a Definition."""
        return (
            "<Definition id='{0}' version='{1}'>".format(
                self.qualified_identifier,
                self.version
            )
        )

    @property
    def path(self):
        """Return path to definition if available.

        :return: Definition :term:`JSON` path or None.

        """
        return self._path

    @property
    def registry_path(self):
        """Return registry path containing the definition if available.

        :return: Registry path or None.

        """
        return self._registry_path

    @property
    def identifier(self):
        """Return definition identifier.

        :return: String value (e.g. "foo").

        .. seealso:: :ref:`definition/identifier`

        """
        return self._data["identifier"]

    @property
    def version(self):
        """Return definition version.

        :return: Instance of :class:`packaging.version.Version` or None.

        :raise: :exc:`wiz.exception.InvalidVersion` if the version is incorrect.

        .. note::

            The value is cached when accessed once to ensure faster access
            afterwards.

        .. seealso:: :ref:`definition/version`

        """
        version = self._data.get("version")

        # Create cache value if necessary.
        if version is not None and self._cache.get("version") is None:
            self._cache["version"] = wiz.utility.get_version(version)

        # Return cached value.
        return self._cache.get("version")

    @property
    def qualified_identifier(self):
        """Return qualified identifier with optional namespace.

        :return: String value (e.g. "namespace::foo").

        """
        if self.namespace is not None:
            return "{}::{}".format(self.namespace, self.identifier)
        return self.identifier

    @property
    def version_identifier(self):
        """Return version identifier.

        :return: String value (e.g. "foo==0.1.0").

        """
        if self.version is not None:
            return "{}=={}".format(self.identifier, self.version)
        return self.identifier

    @property
    def qualified_version_identifier(self):
        """Return qualified version identifier with optional namespace.

        :return: String value (e.g. "namespace::foo==0.1.0").

        """
        if self.namespace is not None:
            return "{}::{}".format(self.namespace, self.version_identifier)
        return self.version_identifier

    @property
    def description(self):
        """Return definition description.

        :return: String value or None.

        .. seealso:: :ref:`definition/description`

        """
        return self._data.get("description")

    @property
    def namespace(self):
        """Return definition namespace.

        :return: String value or None.

        .. seealso:: :ref:`definition/namespace`

        """
        return self._data.get("namespace")

    @property
    def auto_use(self):
        """Return whether definition should be automatically requested.

        :return: Boolean value.

        .. seealso:: :ref:`definition/auto-use`

        """
        return self._data.get("auto-use", False)

    @property
    def disabled(self):
        """Return whether definition is disabled.

        :return: Boolean value.

        .. seealso:: :ref:`definition/disabled`

        """
        return self._data.get("disabled", False)

    @property
    def install_root(self):
        """Return root installation path.

        :return: Directory path or None.

        .. seealso:: :ref:`definition/install_root`

        """
        return self._data.get("install-root")

    @property
    def install_location(self):
        """Return installation path.

        :return: Directory path or None.

        .. seealso:: :ref:`definition/install_location`

        """
        return self._data.get("install-location")

    @property
    def environ(self):
        """Return environment variable mapping.

        :return: Dictionary value.

        .. seealso:: :ref:`definition/environ`

        """
        return self._data.get("environ", {})

    @property
    def command(self):
        """Return command mapping.

        :return: Dictionary value.

        .. seealso:: :ref:`definition/command`

        """
        return self._data.get("command", {})

    @property
    def system(self):
        """Return system requirement mapping.

        :return: Dictionary value.

        .. seealso:: :ref:`definition/system`

        """
        return self._data.get("system", {})

    @property
    def requirements(self):
        """Return list of requirements.

        :return: List of :class:`packaging.requirements.Requirement` instances.

        :raise: :exc:`wiz.exception.InvalidRequirement` if one requirement is
            incorrect.

        .. note::

            The value is cached when accessed once to ensure faster access
            afterwards.

        .. seealso:: :ref:`definition/requirements`

        """
        requirements = self._data.get("requirements")

        # Create cache value if necessary.
        if requirements is not None and self._cache.get("requirements") is None:
            self._cache["requirements"] = [
                wiz.utility.get_requirement(requirement)
                for requirement in requirements
            ]

        # Return cached value.
        return self._cache.get("requirements", [])

    @property
    def conditions(self):
        """Return list of conditions.

        :return: List of :class:`packaging.requirements.Requirement` instances.

        :raise: :exc:`wiz.exception.InvalidRequirement` if one requirement is
            incorrect.

        .. note::

            The value is cached when accessed once to ensure faster access
            afterwards.

        .. seealso:: :ref:`definition/conditions`

        """
        conditions = self._data.get("conditions")

        # Create cache value if necessary.
        if conditions is not None and self._cache.get("conditions") is None:
            self._cache["conditions"] = [
                wiz.utility.get_requirement(condition)
                for condition in conditions
            ]

        # Return cached value.
        return self._cache.get("conditions", [])

    @property
    def variants(self):
        """Return list of conditions.

        :return: List of :class:`Variant` instances.

        .. note::

            The value is cached when accessed once to ensure faster access
            afterwards.

        .. seealso:: :ref:`definition/variants`

        """
        variants = self._data.get("variants")

        # Create cache value if necessary.
        if variants is not None and self._cache.get("variants") is None:
            self._cache["variants"] = [
                Variant(variant, definition_identifier=self.identifier)
                for variant in variants
            ]

        # Return cached value.
        return self._cache.get("variants", [])

    def set(self, element, value):
        """Returns copy of instance with *element* set to *value*.

        :param element: Keyword to add or update in mapping.

        :param value: New value to set as keyword value.

        :return: New updated mapping.

        """
        data = self.data()
        data[element] = value

        return Definition(
            data, path=self._path,
            registry_path=self._registry_path,
            copy_data=False
        )

    def update(self, element, value):
        """Returns copy of instance with *element* mapping updated with *value*.

        :param element: keyword associated to a dictionary.

        :param value: mapping to update *element*  dictionary with.

        :return: New updated mapping.

        :raise: :exc:`ValueError` if *element* is not a dictionary.

        """
        data = self.data()
        data.setdefault(element, {})

        if not isinstance(data[element], dict):
            raise ValueError(
                "Impossible to update '{}' as it is not a "
                "dictionary.".format(element)
            )

        data[element].update(value)

        return Definition(
            data, path=self._path,
            registry_path=self._registry_path,
            copy_data=False
        )

    def extend(self, element, values):
        """Returns copy of instance with *element* list extended with *values*.

        :param element: keyword associated to a list.

        :param values: Values to extend *element* list with.

        :return: New updated mapping.

        :raise: :exc:`ValueError` if *element* is not a list.

        """
        data = self.data()
        data.setdefault(element, [])

        if not isinstance(data[element], list):
            raise ValueError(
                "Impossible to extend '{}' as it is not a list.".format(element)
            )

        data[element].extend(values)

        return Definition(
            data, path=self._path,
            registry_path=self._registry_path,
            copy_data=False
        )

    def insert(self, element, value, index):
        """Returns copy of instance with *value* inserted in *element* list.

        :param element: keyword associated to a list.

        :param value: Value which will be added to the *element* list

        :param index: Index number at which the *value* should be inserted.

        :return: New updated mapping.

        :raise: :exc:`ValueError` if *element* is not a list.

        """
        data = self.data()
        data.setdefault(element, [])

        if not isinstance(data[element], list):
            raise ValueError(
                "Impossible to insert '{}' in '{}' as it is not "
                "a list.".format(value, element)
            )

        data[element].insert(index, value)

        return Definition(
            data, path=self._path,
            registry_path=self._registry_path,
            copy_data=False
        )

    def remove(self, element):
        """Returns copy of instance without *element*.

        :param element: keyword to remove from mapping.

        :return: New updated mapping or "self" if *element* didn't exist in
            mapping.

        """
        data = self.data()
        if element not in data.keys():
            return self

        del data[element]

        return Definition(
            data, path=self._path,
            registry_path=self._registry_path,
            copy_data=False
        )

    def remove_key(self, element, value):
        """Returns copy of instance without key *value* from *element* mapping.

        If *element* mapping is empty after removing *value*, the *element* key
        will be removed.

        :param element: keyword associated to a dictionary.

        :param value: Value to remove from *element* dictionary.

        :return: New updated mapping.

        :raise: :exc:`ValueError` if *element* is not a dictionary.

        """
        data = self.data()
        if element not in data.keys():
            return self

        if not isinstance(data[element], dict):
            raise ValueError(
                "Impossible to remove key from '{}' as it is not a "
                "dictionary.".format(element)
            )

        if value not in data[element].keys():
            return self

        del data[element][value]
        if len(data[element]) == 0:
            del data[element]

        return Definition(
            data, path=self._path,
            registry_path=self._registry_path,
            copy_data=False
        )

    def remove_index(self, element, index):
        """Returns copy of instance without *index* from *element* list.

        If *element* list is empty after removing *index*, the *element* key
        will be removed.

        :param element: keyword associated to a list.

        :param index: Index to remove from *element* list.

        :return: New updated mapping.

        :raise: :exc:`ValueError` if *element* is not a list.

        """
        data = self.data()
        if element not in data.keys():
            return self

        if not isinstance(data[element], list):
            raise ValueError(
                "Impossible to remove index from '{}' as it is not a "
                "list.".format(element)
            )

        if index >= len(data[element]):
            return self

        del data[element][index]
        if len(data[element]) == 0:
            del data[element]

        return Definition(
            data, path=self._path,
            registry_path=self._registry_path,
            copy_data=False
        )

    def data(self, copy_data=True):
        """Return definition data used to created the definition instance.

        :param copy_data: Indicate whether definition data will be copied to
            prevent mutating it. Default is True.

        :return: Definition data mapping.

        """
        if not copy_data:
            return self._data
        return copy.deepcopy(self._data)

    def ordered_data(self, copy_data=True):
        """Return copy of definition data as :class:`collections.OrderedDict`.

        Definition keywords will be sorted as follows:

            1. identifier
            2. version
            3. namespace
            4. description
            5. install-root
            6. install-location
            7. auto-use
            8. disabled
            9. system
            10. command
            11. environ
            12. requirements
            13. conditions
            14. variants

        :ref:`System <definition/system>` keywords will be sorted as follows:

            1. platform
            2. os
            3. arch

        Each :ref:`variant <definition/variants>` mapping will be sorted as
        follows:

            1. identifier
            2. install-location
            3. command
            4. environ
            5. requirements

        :param copy_data: Indicate whether definition data will be copied to
            prevent mutating it. Default is True.

        :return: Instance of :class:`collections.OrderedDict`.

        """
        definition_keywords = [
            "identifier", "version", "namespace", "description",
            "install-root", "install-location", "auto-use", "disabled",
            "system", "command", "environ", "requirements", "conditions",
            "variants"
        ]

        system_keywords = ["platform", "os", "arch"]

        variant_keywords = [
            "identifier", "install-location", "command", "environ",
            "requirements"
        ]

        def _create_ordered_dict(mapping, keywords):
            """Return ordered dictionary from mapping according to keywords.
            """
            content = collections.OrderedDict()

            for keyword in keywords:
                if keyword not in mapping:
                    continue

                if keyword == "system":
                    content[keyword] = _create_ordered_dict(
                        mapping["system"], system_keywords
                    )

                elif keyword == "variants":
                    content[keyword] = [
                        _create_ordered_dict(variant, variant_keywords)
                        for variant in mapping["variants"]
                    ]

                elif isinstance(mapping[keyword], dict):
                    content[keyword] = collections.OrderedDict(
                        sorted(mapping[keyword].items())
                    )

                else:
                    content[keyword] = mapping[keyword]

            return content

        return _create_ordered_dict(
            self.data(copy_data=copy_data), definition_keywords
        )

    def encode(self):
        """Return serialized definition data.

        :class:`collections.OrderedDict` instance as returned by
        :meth:`ordered_data` is being used.

        :return: Serialized mapping.

        """
        return json.dumps(
            self.ordered_data(),
            indent=4,
            separators=(",", ": "),
            ensure_ascii=False
        )


class Variant(object):
    """Definition variant object."""

    def __init__(self, data, definition_identifier):
        """Initialize definition variant.

        :param data: Variant data definition mapping.

        :param definition_identifier: Identifier of the definition containing
            the variant data.

        .. warning::

            "requirements" values will not get validated when constructing the
            instance for performance reason. Therefore, accessing this value
            could raise an error when data is incorrect::

                >>> variant = Variant(
                ...     {
                ...         "identifier": "variant1",
                ...         "requirements": ["!!!"],
                ...     },
                ...     definition_identifier="foo"
                ... )
                >>> print(variant.requirements)

                InvalidRequirement: The requirement '!!!' is incorrect

        .. seealso:: :ref:`definition/variants`

        """
        self._data = data
        self._definition_identifier = definition_identifier

        # Store values that needs to be constructed.
        self._cache = {}

    @property
    def identifier(self):
        """Return variant identifier.

        :return: String value (e.g. "variant1").

        """
        return self._data["identifier"]

    @property
    def definition_identifier(self):
        """Return definition identifier.

        :return: String value (e.g. "foo").

        """
        return self._definition_identifier

    @property
    def install_location(self):
        """Return installation path.

        :return: Directory path or None.

        .. seealso:: :ref:`definition/install_location`

        """
        return self._data.get("install-location")

    @property
    def environ(self):
        """Return environment variable mapping.

        :return: Dictionary value.

        .. seealso:: :ref:`definition/environ`

        """
        return self._data.get("environ", {})

    @property
    def command(self):
        """Return command mapping.

        :return: Dictionary value.

        .. seealso:: :ref:`definition/command`

        """
        return self._data.get("command", {})

    @property
    def requirements(self):
        """Return list of requirements.

        :return: List of :class:`packaging.requirements.Requirement` instances.

        :raise: :exc:`wiz.exception.InvalidRequirement` if one requirement is
            incorrect.

        .. note::

            The value is cached when accessed once to ensure faster access
            afterwards.

        .. seealso:: :ref:`definition/requirements`

        """
        requirements = self._data.get("requirements")

        # Create cache value if necessary.
        if requirements is not None and self._cache.get("requirements") is None:
            self._cache["requirements"] = [
                wiz.utility.get_requirement(requirement)
                for requirement in requirements
            ]

        # Return cached value.
        return self._cache.get("requirements", [])

    def data(self, copy_data=True):
        """Return variant data used to created the variant instance.

        :param copy_data: Indicate whether definition data will be copied to
            prevent mutating it. Default is True.

        :return: Variant data mapping.

        """
        if not copy_data:
            return self._data
        return copy.deepcopy(self._data)
